/*
 * large_sliced_coo_kernel.h
 *
 *  Created on: Dec 13, 2010
 *      Author: hans
 */

#ifndef LARGE_SLICED_COO_KERNEL_H_
#define LARGE_SLICED_COO_KERNEL_H_

#include <stdint.h>
#include "../util/util.h"
#include "../util/textures.h"
#include "../util/xor.h"

namespace cuda_bwc{

__inline__ __host__ __device__ uint16_t combine(uint8_t first, uint8_t second){
	return first | (second<<8);
}

template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE>
__global__ void
large_sliced_coo_kernel_32(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t numGroups,
                const uint32_t groupSize,
                const uint32_t * cols,
                const uint8_t * rows,
                const ValueType *V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
    const uint32_t thread_lane = threadIdx.x & 31;
    const uint32_t row_lane = threadIdx.x/32;
    __shared__ ValueType sdata[NUM_ROWS_PER_SLICE][32];
 
    const uint32_t packNo=blockIdx.x;   
    const uint32_t limit = ( (packNo==numPacks-1)?((num_rows-1)%NUM_ROWS_PER_SLICE)+1:NUM_ROWS_PER_SLICE );


        syncthreads();
        for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK/32){
            sdata[i][thread_lane] = 0;
        }
        syncthreads();

        uint32_t groupVal = 0;
        const uint32_t base_offset = packNo*numGroups;
        int32_t begin = offsets[base_offset];
        int32_t end;

        for(uint32_t i=1; i<=numGroups; i++){
        	end = offsets[base_offset+i];
        	for(int32_t index=begin+threadIdx.x; index<end; index+=THREADS_PER_BLOCK){
        		const uint32_t col = cols[index] + groupVal;
				const uint8_t row = rows[index];
				ValueType input = fetch_x(col, x) * V[index];
                
				//atomic_xor(&sdata[row][thread_lane], input);
                atomic_add(&sdata[row][thread_lane], input);
        	}
        	groupVal += groupSize;
        	begin = end;
        }

        syncthreads();

        for(int j=16; j>=1; j>>=1){
            for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK/32){
                if(thread_lane < j){
                    sdata[i][(thread_lane+i)&31] += sdata[i][(thread_lane+i+j)&31];
                }
            }
            syncthreads();
        }

    const uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;

    for(uint32_t r = threadIdx.x; r < limit; r+=THREADS_PER_BLOCK){
       y[actualRow+r] = sdata[r][thread_lane];
    }

}

template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE>
__global__ void
large_sliced_coo_kernel_64_128_256(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t numGroups,
                const uint32_t groupSize,
                const uint8_t * coo,
                const double  * V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
    const uint32_t thread_lane = threadIdx.x & 31;
    const uint32_t row_lane = threadIdx.x/32;
    const uint32_t REPEAT = sizeof(ValueType)/sizeof(uint32_t);
    __shared__ uint32_t sdata[NUM_ROWS_PER_SLICE][REPEAT][32];

    uint32_t packNo=blockIdx.x;
	uint32_t limit = NUM_ROWS_PER_SLICE;
//	bool swap = false;
	
//    for(packNo=blockIdx.x,base=0; packNo<numPacks; ){
    	if(packNo == numPacks-1){
    		limit = ((num_rows-1)%NUM_ROWS_PER_SLICE)+1;
    	}

    	syncthreads();
        for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK/32){
        	scattered_write<uint32_t,REPEAT>(&sdata[i][0][thread_lane], 32, 0);
        }
        syncthreads();
        
        uint32_t groupVal = 0;
        const uint32_t base_offset = packNo*numGroups;
        int32_t begin = offsets[base_offset];
        int32_t end;

        for(uint32_t i=1; i<=numGroups; i++){
        	end = offsets[base_offset+i];
        	for(int32_t index=begin+threadIdx.x*3; index<end; index+=THREADS_PER_BLOCK*3){
        		uint32_t col = combine(coo[index], coo[index+1]) + groupVal;
				uint8_t row = coo[index+2];
				ValueType input = fetch_x(col, x);
				atomic_xor(&sdata[row][0][thread_lane], input, 32);
        	}
        	groupVal += groupSize;
        	begin = end;
        }

        syncthreads();

        for(uint32_t j=16; j>=1; j>>=1){
            for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK/32){
                if(thread_lane < j){
                	XOR<REPEAT>(&sdata[i][0][(thread_lane+i)&(31)], &sdata[i][0][(thread_lane+i+j)&(31)], 32);
                }
            }
            syncthreads();
        }

        uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;

        for(uint32_t id=threadIdx.x; id<limit*REPEAT; id+=THREADS_PER_BLOCK){
            reinterpret_cast<uint32_t*>(&y[actualRow+id/REPEAT])[id&(REPEAT-1)] = sdata[id/REPEAT][id&(REPEAT-1)][(id/REPEAT)&31];
        }
        
//        base+=gridDim.x;
//        if(swap){
//        	packNo = base + gridDim.x - 1 - blockIdx.x;
//        }else{
//        	packNo = base + blockIdx.x;
//        }
//        swap = !swap;
//
//    }
}

}
#endif /* LARGE_SLICED_COO_KERNEL_H_ */
