/*
 * sliced_ellpack.h
 *
 *  Created on: Dec 15, 2010
 *      Author: hans
 */

#ifndef SLICED_ELLPACK_H_
#define SLICED_ELLPACK_H_

namespace cuda_bwc{

#include <sstream>
#include <stdint.h>
#include "binary_matrix.h"
#include "sliced_ellpack_kernel.cuh"

template <typename T>
class SlicedEllpack : public BinaryMatrix<T>{
    public:
        SlicedEllpack() : BinaryMatrix<T>(0,0){};
        SlicedEllpack(uint32_t num_rows, uint32_t num_cols) :
            BinaryMatrix<T>(num_rows,num_cols){}
        ~SlicedEllpack();
        void readMatrix(MatrixInput &in);
        void writeMatrix(std::ostream &out);
        void readCache(std::istream &in);
        void buildCache(std::ostream &out);
        std::string getCacheName();
        uint64_t memoryUsage();
        void multiply(const T *v, T *r);
        uint32_t granularity();
        uint32_t nonZeroes();

    protected:
        void transferToDevice();

    private:
        thrust::device_vector<uint32_t> dv_c_index;
        thrust::device_vector<uint32_t> dv_slice_start;
        thrust::device_vector<T> dv_value;

        thrust::host_vector<uint32_t> c_index;
        thrust::host_vector<uint32_t> slice_start;
        thrust::host_vector<T> value;
        

        uint32_t non_zeroes;
        static const uint32_t THREADS_PER_BLOCK = 512;
        static const uint32_t HEIGHT = THREADS_PER_BLOCK;
        void addElem(uint32_t elm, T value=0);
};

template <typename T>
SlicedEllpack<T>::~SlicedEllpack() {
}

template <typename T>
std::string SlicedEllpack<T>::getCacheName(){
    std::stringstream ss;
    ss << "-sle-" << this->num_rows << "-" << this->num_cols << ".bin";
    return ss.str();
}

template <typename T>
void SlicedEllpack<T>::addElem(uint32_t elem, T v){
//    unsigned char* c = reinterpret_cast<unsigned char*>(&elem);
//    c_index.push_back(c[0]);
//    c_index.push_back(c[1]);
//    c_index.push_back(c[2]);
    c_index.push_back(elem);  
    value.push_back(v);
}


template <typename T>
void SlicedEllpack<T>::readMatrix(MatrixInput &in){
    vector< pair < uint32_t, T> > vt[HEIGHT];
    uint32_t nz = 0;
    uint32_t numEntries = 0;
    slice_start.push_back(0);
    for (uint32_t i=0; i<this->num_rows;){
        uint32_t limit;
        uint32_t height;
        if(i+HEIGHT < this->num_rows){
            limit = i+HEIGHT;
            height = HEIGHT;
        }else{
            limit = this->num_rows;
            height = this->num_rows - i;
        }
        uint32_t max_row = 0;
        for(;i<limit; i++){
            uint32_t nc;
            in >> nc;
            nz += nc;
            if (nc > max_row) max_row = nc;
            uint32_t p = i%HEIGHT;
            vt[p].clear();
            for (int j=0;j<nc;j++){
                uint32_t value;
                in >> value;
                T v = (T) in.getValue();
                vt[p].push_back( make_pair( value, v ) );
            }
        }
        
        numEntries+=max_row*height;
        slice_start.push_back(numEntries);
        for (int jj = 0; jj < max_row; jj++) {
            for (int j = 0; j < height; j++) {
                if (jj < max_row - vt[j].size()) {
                    addElem(0);
                } else {
                    uint32_t temp = jj - (max_row - vt[j].size());
                    addElem(vt[j][temp].first+1, vt[j][temp].second );
                }
            }
        }
        max_row = 0;
    }
    this->non_zeroes = nz;
    transferToDevice();
}

template <typename T>
void SlicedEllpack<T>:: writeMatrix(std::ostream &out){
}

template <typename T>
void SlicedEllpack<T>::readCache(std::istream &in){
    DataInputStream dis(in);
    dis >> this->num_rows >> this->num_cols >> non_zeroes;
    dis.readVector(c_index);
    dis.readVector(slice_start);
    dis.readVector(value);
    transferToDevice();
}

template <typename T>
void SlicedEllpack<T>::buildCache(std::ostream &out){
    DataOutputStream dos(out);
    dos << this->num_rows << this->num_cols << non_zeroes;
    dos.writeVector(c_index);
    dos.writeVector(slice_start);
    dos.writeVector(value);
}

template <typename T>
void SlicedEllpack<T>::transferToDevice(){
    dv_c_index.assign(c_index.begin(), c_index.end());
    dv_slice_start.assign(slice_start.begin(), slice_start.end());
    dv_value.assign( value.begin() , value.end() );
}

template <typename T>
uint64_t SlicedEllpack<T>::memoryUsage(){
    return c_index.size()*sizeof(uint8_t) + slice_start.size() * sizeof(uint32_t);
}

template <typename T>
void SlicedEllpack<T>::multiply(const T *v, T *r){
//	static const uint32_t NUM_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(
//			sliced_ellpack_kernel<T, THREADS_PER_BLOCK> , THREADS_PER_BLOCK, (size_t) 0);

    bind_x(v);
	bind_Ap(thrust::raw_pointer_cast(&dv_slice_start[0]));
	sliced_ellpack_kernel<T, THREADS_PER_BLOCK> <<<divideAndCeil(this->num_rows, HEIGHT), THREADS_PER_BLOCK>>>
	                (this->num_rows,
	                thrust::raw_pointer_cast(&dv_slice_start[0]),
	                thrust::raw_pointer_cast(&dv_c_index[0]),
                    thrust::raw_pointer_cast(&dv_value[0]),
	                v, r);
	unbind_Ap(thrust::raw_pointer_cast(&dv_slice_start[0]));
    unbind_x(v);
}

template <typename T>
uint32_t SlicedEllpack<T>::granularity(){
	return HEIGHT;
}


template <typename T>
uint32_t SlicedEllpack<T>::nonZeroes(){
	return non_zeroes;
}


}

#endif /* SLICED_ELLPACK_H_ */
