/*
 * small_sliced_coo_kernel.h
 *
 *  Created on: Dec 14, 2010
 *      Author: hans
 */

#ifndef SMALL_SLICED_COO_KERNEL_H_
#define SMALL_SLICED_COO_KERNEL_H_

namespace cuda_bwc{

template <typename ValueType, const uint32_t NUM_ROWS_PER_SLICE, const uint32_t THREADS_PER_BLOCK>
__global__ void
small_sliced_coo_kernel_32(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t * cols,
                const uint8_t * rows,
                const ValueType * V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
    volatile __shared__ ValueType sdata[NUM_ROWS_PER_SLICE][THREADS_PER_BLOCK];

    const int packNo = blockIdx.x;
    const int limit = ( (packNo==numPacks-1)?((num_rows-1)%NUM_ROWS_PER_SLICE)+1:NUM_ROWS_PER_SLICE );

    const uint32_t begin = offsets[packNo];
    const uint32_t end = offsets[packNo+1];
    for(int i=0; i<limit; i++)
        sdata[i][threadIdx.x] = 0;
    
    syncthreads();
    
    for(uint32_t index=begin+threadIdx.x; index<end; index+=THREADS_PER_BLOCK){
        const uint32_t col = cols[index]; 
        const uint8_t row = rows[index];
        ValueType input = fetch_x(col, x) * V[index];

        sdata[row][threadIdx.x] += input;
    }
    syncthreads();
        
    for(int j=THREADS_PER_BLOCK/2; j>32; j>>=1){
        for(int i=0; i<limit; i++){
            if(threadIdx.x < j){
                sdata[i][(threadIdx.x+i)&(THREADS_PER_BLOCK-1)] += sdata[i][(threadIdx.x+i+j)&(THREADS_PER_BLOCK-1)];
            }
        }    
        syncthreads();
    }

    for (int i=0;i<limit;i++) {
        const int tid = (threadIdx.x + i) & (THREADS_PER_BLOCK-1);
        if (threadIdx.x<32) sdata[i][tid] += sdata[i][(tid+32) & (THREADS_PER_BLOCK-1)];
        if (threadIdx.x<16) sdata[i][tid] += sdata[i][(tid+16) & (THREADS_PER_BLOCK-1)];
        if (threadIdx.x<8) sdata[i][tid] += sdata[i][(tid+8) & (THREADS_PER_BLOCK-1)];
        if (threadIdx.x<4) sdata[i][tid] += sdata[i][(tid+4) & (THREADS_PER_BLOCK-1)];
        if (threadIdx.x<2) sdata[i][tid] += sdata[i][(tid+2) & (THREADS_PER_BLOCK-1)];
        if (threadIdx.x<1) sdata[i][tid] += sdata[i][(tid+1) & (THREADS_PER_BLOCK-1)];
    }

    const uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;
    syncthreads();

    if(threadIdx.x < limit){
        y[actualRow+threadIdx.x] = sdata[threadIdx.x][threadIdx.x];
    }

}

template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE>
__global__ void
small_sliced_coo_kernel_64_128_256(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t numGroups,
                const uint32_t groupSize,
                const uint8_t * coo,
                const ValueType * V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
    const uint32_t REPEAT = sizeof(ValueType)/sizeof(uint64_t);
	const uint32_t JUMP = THREADS_PER_BLOCK*3;
    __shared__ uint64_t sdata[NUM_ROWS_PER_SLICE][REPEAT][THREADS_PER_BLOCK];

    uint32_t packNo=blockIdx.x;
	uint32_t limit = NUM_ROWS_PER_SLICE;
//	bool swap = false;
	
//    for(packNo=blockIdx.x,base=0; packNo<numPacks;){
    	if(packNo == numPacks-1){
    		limit = ((num_rows-1)%NUM_ROWS_PER_SLICE)+1;
    	}
        syncthreads();
        for(uint32_t i=0; i<NUM_ROWS_PER_SLICE; i++){
            scattered_write<uint64_t,REPEAT>(&sdata[i][0][threadIdx.x], THREADS_PER_BLOCK, 0);
        }
        syncthreads();

        uint32_t groupVal = 0;
        const uint32_t base_offset = packNo*numGroups;
        int32_t begin = offsets[base_offset];
        int32_t end;

        for(uint32_t i=1; i<=numGroups; i++){
        	end = offsets[base_offset+i];
        	int32_t index=begin+threadIdx.x*3;
        	for(; index<end; index+=JUMP){
        		uint32_t col = combine(coo[index], coo[index+1]) + groupVal;
				uint8_t row = coo[index+2];
				ValueType input = fetch_x(col, x);
				XOR(&sdata[row][0][threadIdx.x], THREADS_PER_BLOCK, input);
        	}
        	
        	groupVal += groupSize;
        	begin = end;
        }

        syncthreads();

        for(unsigned int i=0; i<limit; i++){
            for(unsigned int j=THREADS_PER_BLOCK/2; j>=1; j>>=1){
                if(threadIdx.x < j){
                    XOR<REPEAT>(&sdata[i][0][(threadIdx.x+i)&(THREADS_PER_BLOCK-1)], &sdata[i][0][(threadIdx.x+i+j)&(THREADS_PER_BLOCK-1)], THREADS_PER_BLOCK);
                }
                syncthreads();
            }
        }

        uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;

        if(threadIdx.x < limit*REPEAT){
            reinterpret_cast<uint64_t*>(&y[actualRow+threadIdx.x/REPEAT])[threadIdx.x&(REPEAT-1)] = sdata[threadIdx.x/REPEAT][threadIdx.x&(REPEAT-1)][threadIdx.x/REPEAT];
        }
        
//        base+=gridDim.x;
//        if(swap){
//        	packNo = base + gridDim.x - 1 - blockIdx.x;
//        }else{
//        	packNo = base + blockIdx.x;
//        }
//        swap = !swap;
//    }
}

}

#endif /* SMALL_SLICED_COO_KERNEL_H_ */
